# Copyright 2025 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import pathlib
import stat
import subprocess
import time
from abc import ABC, abstractmethod
from typing import Optional, Dict, Any

from planet_auth.auth_exception import AuthException
from planet_auth.util import auth_logger


ObjectStorageProvider_KeyType = pathlib.Path
"""
Key type for object storage.  Paths are currently used in part because of
the history of the class.  Paths need not be real files on disk.
It is up to the implementation to decide if a real file should be used,
or another storage mechanism should be used.
"""


class ObjectStorageProvider(ABC):
    """
    Abstract interface used to persist objects to storage.
    """

    # TODO: do we need a concept of "list" for the storage provider?
    #     The CLI's "profiles list" would be the only consumer at this time,
    #     and the CLI does not support custom storage providers yet.

    @abstractmethod
    def load_obj(self, key: ObjectStorageProvider_KeyType) -> dict:
        """
        Read an object from storage
        :return:
        """

    @abstractmethod
    def save_obj(self, key: ObjectStorageProvider_KeyType, data: dict) -> None:
        """
        Write an object to storage
        :return:
        """

    @abstractmethod
    def obj_exists(self, key: ObjectStorageProvider_KeyType) -> bool:
        """
        Check whether a given object exists in storage.
        """

    @staticmethod
    def _default_storage_provider():
        # Always create JIT to better handle cases where runtime changes to
        # the env may impact behavior (like code that sets HOME in the env).
        # Not great for performance.  Probably an edge case, but one unit
        # tests depend on.  There should be a better way.
        return _SOPSAwareFilesystemObjectStorageProvider()
        # if not ObjectStorageProvider._default_storage_provider_impl:
        #     ObjectStorageProvider._default_storage_provider_impl = _SOPSAwareFilesystemObjectStorageProvider()
        # return ObjectStorageProvider._default_storage_provider_impl


# TODO: Should we also provide a reference storage provider implementation for keyring?
#    We would have to work out how we want independent installations
#    to interact, if not through ~/.planet/.  How would QGIS interact with the CLI
#    to manage a session, for example?
#    (See https://pypi.org/project/keyring/)
class _SOPSAwareFilesystemObjectStorageProvider(ObjectStorageProvider):
    """
    Storage provider geared around backing a single object in a single file
    with paths take from the root of the local file system.
    """

    def __init__(self, root: Optional[pathlib.Path] = None):
        if root:
            self._storage_root = root
        else:
            # self._storage_root = pathlib.Path.home() / ".planet"
            # self._storage_root = pathlib.Path("/")
            self._storage_root = pathlib.Path.home()

    def _obj_filepath(self, obj_key):
        if obj_key.is_absolute():
            # obj_path = self._storage_root / obj_key.relative_to("/")
            obj_path = obj_key
        else:
            obj_path = self._storage_root / obj_key
        return obj_path

    @staticmethod
    def _is_sops_path(file_path):
        # TODO: Could be ".json.sops", or ".sops.json", depending on file
        #   level or field level encryption, respectively.  We currently
        #   only look for and support field level encryption in json
        #   files with a ".sops.json" suffix.
        return bool(file_path.suffixes == [".sops", ".json"])

    @staticmethod
    def _read_json(file_path: pathlib.Path):
        auth_logger.debug(msg="Loading JSON data from file {}".format(file_path))
        with open(file_path, mode="r", encoding="UTF-8") as file_r:
            return json.load(file_r)

    @staticmethod
    def _read_json_sops(file_path: pathlib.Path):
        auth_logger.debug(msg="Loading JSON data from SOPS encrypted file {}".format(file_path))
        data_b = subprocess.check_output(["sops", "-d", file_path])
        return json.loads(data_b)

    @staticmethod
    def _write_json(file_path: pathlib.Path, data: dict):
        auth_logger.debug(msg="Writing JSON data to file {}".format(file_path))
        with open(file_path, mode="w", encoding="UTF-8") as file_w:
            os.chmod(file_path, stat.S_IREAD | stat.S_IWRITE)
            _no_none_data = {key: value for key, value in data.items() if value is not None}
            file_w.write(json.dumps(_no_none_data, indent=2, sort_keys=True))

    @staticmethod
    def _write_json_sops(file_path: pathlib.Path, data: dict):
        # TODO: It would be nice to only encrypt the fields we need to.
        #       It would be a better user experience.
        #
        # Writing directly seems to blow up. I guess we have to write clear text,
        # then encrypt in place?
        # with io.StringIO(json.dumps(data)) as data_f:
        #     subprocess.check_call(
        #         ['sops', '-e', '--input-type', 'json', '--output-type',
        #          'json', '--output', file_path, '/dev/stdin'],
        #         stdin=data_f)
        auth_logger.debug(msg="Writing JSON data to SOPS encrypted file {}".format(file_path))
        _SOPSAwareFilesystemObjectStorageProvider._write_json(file_path, data)
        subprocess.check_call(["sops", "-e", "--input-type", "json", "--output-type", "json", "-i", file_path])

    @staticmethod
    def _load_file(file_path: pathlib.Path) -> dict:
        if _SOPSAwareFilesystemObjectStorageProvider._is_sops_path(file_path):
            new_data = _SOPSAwareFilesystemObjectStorageProvider._read_json_sops(file_path)
        else:
            new_data = _SOPSAwareFilesystemObjectStorageProvider._read_json(file_path)

        return new_data

    @staticmethod
    def _save_file(file_path: pathlib.Path, data: dict):
        file_path.parent.mkdir(parents=True, exist_ok=True)
        if _SOPSAwareFilesystemObjectStorageProvider._is_sops_path(file_path):
            _SOPSAwareFilesystemObjectStorageProvider._write_json_sops(file_path, data)
        else:
            _SOPSAwareFilesystemObjectStorageProvider._write_json(file_path, data)

    def load_obj(self, key: ObjectStorageProvider_KeyType) -> dict:
        obj_filepath = self._obj_filepath(key)
        return self._load_file(file_path=obj_filepath)

    def save_obj(self, key: ObjectStorageProvider_KeyType, data: dict) -> None:
        obj_filepath = self._obj_filepath(key)
        self._save_file(file_path=obj_filepath, data=data)

    def obj_exists(self, key: ObjectStorageProvider_KeyType) -> bool:
        obj_filepath = self._obj_filepath(key)
        return obj_filepath.exists()


class FileBackedJsonObjectException(AuthException):
    """
    Exception indicating a problem with a file backed json object.
    """

    def __init__(self, file_path=None, **kwargs):
        super().__init__(**kwargs)
        self._file_path = file_path

    def set_file_path(self, file_path):
        self._file_path = file_path

    def __str__(self):
        if self._file_path:
            return "{} (File: {})".format(super().__str__(), self._file_path)
        else:
            return super().__str__()


class InvalidDataException(FileBackedJsonObjectException):
    pass


# TODO: rename/generalize as a storage backed JSON object. Since this was
#   originally written, we have introduced the concept of custom storage
#   providers, and storage may not always be file backed.
class FileBackedJsonObject:
    """
    A storage backed json object for storing information. Base class provides
    lazy loading and validation before saving or setting the data.  Derived
    classes should provide type specific validation and convenience data
    accessors. The intent is that this provides a way to have an object
    backed by storage, and rails are in place to prevent invalid data from
    being saved or used.

    Data is allowed to be initialized as None so that JIT load() use cases
    can be supported with the data not being read until needed.  Care
    should be taken when constructing from an in-memory data value, as
    leaf class data validation is not performed during construction.

    The None data structure is treated as a special case, and means that the
    class has been constructed, but the data has never been set.  Once data has
    been set, either by calling set_data() or by loading data from a storage,
    it is only permitted to be set to valid values.

    Saving to the backing is also optional. Uses in this way, this class
    behaves as a validating in-memory json object holder.  To disable
    saving to storage, construct objects with a None file_path.

    The default storage provider will use the file system for storage.
    When loading from file storage, the file may be SOPS encrypted.
    When using SOPS encryption, the file should have a ".sops.json" suffix.
    """

    # TODO: Consider a shift to a schema framework for validation?
    #       E.g. schematics Model.

    def __init__(
        self,
        data: Optional[Dict[str, Any]] = None,
        file_path: Optional[pathlib.Path] = None,
        storage_provider: Optional[ObjectStorageProvider] = None,
    ):
        # Derived classes should not do anything to make the base class's
        # self._data non-falsey based on empty init to keep JIT lazy_load()
        # use cases from file backed object working as expected.
        self._file_path = pathlib.Path(file_path) if file_path else None
        self._load_time = 0
        self._data: Dict[str, Any] = data  # type: ignore
        if data:
            # We used to do a self.check_data(data) to try and prevent
            # invalid data from being constructed, but that's dubious in
            # a base class when the goal of check_data() was to enforce leaf
            # class requirements. The leaf classes are likely not fully
            # constructed at this point.
            # self.check_data(data)
            self._load_time = int(time.time())

        if storage_provider:
            self._object_storage_provider = storage_provider
        else:
            self._object_storage_provider = ObjectStorageProvider._default_storage_provider()

    def __json_pretty_dumps__(self):
        # This function is provided so json.dumps can display
        # versions of the object that are different from what
        # we want to save to storage.
        # This is just to work with pretty-print object dumper in
        # the plauth CLI.  We prefer to not strip the null/None
        # entries when we dump to a file so the act as guideposts
        # to folks manually editing the file. And, when we dump to
        # a file we do not write the path of the file into the
        # file - it's self-evident, and would be over constrained.
        def _del_none(d):
            for key, value in list(d.items()):
                if value is None:
                    del d[key]
                elif isinstance(value, dict):
                    _del_none(value)
            return d

        json_dumps = self._data.copy()
        if self._file_path:
            json_dumps["_file_path"] = str(self._file_path)
        _del_none(json_dumps)
        return json_dumps

    def path(self) -> Optional[pathlib.Path]:
        """
        Retrieve the currently configured path for saved data.
        """
        return self._file_path

    def set_path(self, file_path):
        """
        Update storage path for the object.
        """
        self._file_path = pathlib.Path(file_path) if file_path else None

    def set_storage_provider(self, storage_provider: Optional[ObjectStorageProvider] = None):
        """
        Update storage provider for the object.
        """
        if storage_provider:
            self._object_storage_provider = storage_provider
        else:
            self._object_storage_provider = ObjectStorageProvider._default_storage_provider()

    def storage_provider(self) -> ObjectStorageProvider:
        """
        Retrieve the currently configured storage provider for saved data.
        """
        return self._object_storage_provider

    def is_persisted_to_storage(self) -> bool:
        """
        Check if the object exists in storage.  This does not require or cause the
        object to be loaded.  This does not check that the version in storage
        is the most up-to-date.
        """
        if self._file_path:
            return self._object_storage_provider.obj_exists(self._file_path)
        else:
            return False

    def data(self):
        """
        Retrieve the current data that is in memory. This method will
        not load the data from storage.
        """
        return self._data

    def update_data(self, sparse_update_data):
        """
        Update the data set with the provided sparse update.
        Updates that result in a data-set that will not pass
        check_data() will be rejected, and the data will be unchanged.
        """
        old_data = self._data
        if old_data:
            new_data = old_data.copy()
            new_data.update(sparse_update_data)
        else:
            new_data = sparse_update_data
        self.set_data(new_data)

    def set_data(self, data):
        """
        Set the current in memory data. The data will be checked for validity
        before in memory values are set.  Invalid data will result in an exception
        being thrown and no change being made to the in memory object.
        """
        self.check_data(data)
        self._data = data.copy()
        self._load_time = int(time.time())

    def check_data(self, data):
        """
        Check that the provided data is valid.  Throws an exception if the
        data is not valid. This allows the base class to refuse to store or
        use data, while leaving it to child classes to know what constitutes
        "valid". Child classes should raise FileBackedJsonObjectException
        exception.

        Child classes should override this method as required to do more
        application specific data integrity checks. They should also
        call validation on their base classes so that all layers of
        validation are performed.

        The base assertion is that the data structure has been set.
        It may be empty, but may not be None.  The None data structure
        is treated as a special case, and means that the class has been
        constructed, but the data has never been set.  Once data has
        been set, it is only permitted to be set to valid values.
        The intent of this behavior is to support JIT loading semantics.
        """
        # Allow empty, but not None
        if not data:
            if data == {}:
                return
            raise InvalidDataException(
                message="None is not valid data for {} type, backed by file '{}'".format(
                    self.__class__.__name__, self._file_path
                )
            )

    def check(self):
        """
        Run check_data against our current state.  An exception will be thrown
        if the current data is found to be invalid.  Subclasses should not
        need to override this method, as they are expected to implement
        check_data().  This method will not perform a load() before checking
        the data. That is considered an application responsibility.
        """
        try:
            self.check_data(self._data)
        except FileBackedJsonObjectException as e:
            if self._file_path:
                e.set_file_path(self._file_path)
            raise e

    def save(self):
        """
        Save the data to storage.  If the data is invalid according
        to the child class's check_data() method, the data will
        not be saved and an error will be thrown. The intent in
        this design is to never save known bad data.

        If the path has not been set, nothing will be saved
        to storage, and this will silently succeed. This is to
        allow for transparent in memory only use cases.
        """
        self.check_data(self._data)

        if not self._file_path:
            # We now allow for in memory operation
            # raise FileBackedJsonObjectException(message="Cannot save data to file. File path is not set.")
            return

        self._object_storage_provider.save_obj(self._file_path, self._data)
        self._load_time = int(time.time())

    def is_loaded(self) -> bool:
        return bool(self._data)

    def load(self):
        """
        Force a data load from storage.  If the file path has not been set,
        this will be a no-op to allow for in memory operations.

        If the data loaded from storage is invalid, an exception will be thrown,
        and the currently held data will be left unchanged.  When in memory
        data is invalid and no file path has been set, an error IS NOT
        thrown - there would be nothing to fall back to. In memory users
        should expect to call check() or check_data() on their own.
        """
        if not self._file_path:
            # raise FileBackedJsonObjectException(message="Cannot load data from file. File path is not set.")
            return  # we now allow in memory operation.  Should we raise an error if the current data is invalid?

        new_data = self._object_storage_provider.load_obj(self._file_path)
        self.check_data(new_data)
        self._data = new_data
        self._load_time = int(time.time())

    def lazy_load(self):
        """
        Lazy load the data from storage.

        If the data has already been set, whether by an the constructor, an
        explicit set_data(), or a previous load from storage, no attempt is made
        to refresh the data.  Object will behave as an in memory object.

        If the data is not set, it will attempt to load the data from storage.
        An error will be thrown if the loaded data is invalid, the file has
        not been set, or the file has been set to a non-existent path.

        For updating the data from storage even if an in memory copy exists,
        see lazy_reload().

        Users may always force a load at any time by calling load()
        """
        if not self.is_loaded():
            self.load()

    def lazy_reload(self):
        """
        Lazy reload the data from storage.

        If the data is set, a reload will be attempted if the data on storage
        appears to be newer if a path is set.  if a path is not set, no
        attempt will be made to load the data.

        If the data is not set, an error will be thrown if the loaded data is
        invalid, the file has not been set, or the file has been set to a
        non-existent path.
        """
        if not self.is_loaded():
            # No data, behave like load()
            self.load()
            return

        if not self._file_path:
            # Have data. No path. Continue with in memory value.
            return

        if int(self._file_path.stat().st_mtime) > self._load_time:
            self.load()

    def lazy_get(self, field):
        """
        Lazy load the data and retrieve the requested field.
        """
        self.lazy_load()
        if self._data:
            return self._data.get(field, None)
        else:
            return None

    # This is probably a bad idea, since it would encourage development
    # that makes it more likely that data may change in between get()'s
    # of multiple fields in the same object, leading to inconsistency.
    # Only the application knows what transaction boundaries are.
    # def lazy_reload_get(self, field):
    #    self.lazy_reload()
    #    if self._data:
    #        return self._data.get(field)
    #    return None



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import sys
from collections import OrderedDict
from prompt_toolkit.shortcuts import input_dialog, radiolist_dialog

from planet_auth import AuthClient, AuthClientConfig, AuthException
from planet_auth.logging.auth_logger import getAuthLogger
from planet_auth.constants import (
    AUTH_CONFIG_FILE_PLAIN,
    AUTH_CONFIG_FILE_SOPS,
)

from planet_auth_utils.profile import Profile
from planet_auth_utils.builtins import Builtins
from planet_auth_utils.plauth_factory import PlanetAuthFactory
from planet_auth_utils.plauth_user_config import PlanetAuthUserConfig
from planet_auth_utils.constants import EnvironmentVariables
from .options import opt_long, opt_sops
from .util import recast_exceptions_to_click, print_obj

auth_logger = getAuthLogger()


def _handle_canceled():
    print("Canceled")
    sys.exit(1)


def _dialogue_choose_auth_client_type():
    choices = []
    for _, config_type in AuthClientConfig._get_typename_map().items():
        client_type = AuthClient._get_type_map().get(config_type)
        client_display_name = config_type.meta().get("display_name") or client_type.__name__
        client_description = config_type.meta().get("description")
        choices.append(([config_type, client_type], "{:40} - {}".format(client_display_name, client_description)))

    return (
        radiolist_dialog(
            title="Authentication Profile Creation",
            text="Select the auth client type.\n"
            "The auth client type determines how the software will interact with authentication and API"
            " services to make requests as the authenticated user.",
            values=choices,
        ).run()
        or _handle_canceled()
    )


def _dialogue_choose_auth_profile():
    all_profile_names = Builtins.builtin_profile_names() + Profile.list_on_disk_profiles()
    choices = []
    for profile_name in all_profile_names:
        choices.append((profile_name, f"{profile_name}"))
    return (
        radiolist_dialog(
            title="Authentication Builtins",
            text="Select the Authentication Profile to use",
            values=choices,
        ).run()
        or _handle_canceled()
    )


def _dialogue_enter_auth_profile_name():
    # TODO:
    #   - Check for collisions with existing profile or built in profiles.
    return (
        input_dialog(
            title="Auth Profile Creation",
            text="Select a name for the new client profile."
            "  Profile names should be legal file system names, and will be normalized.",
        ).run()
        or _handle_canceled()
    )


@click.group("profile", invoke_without_command=True)
@click.pass_context
def cmd_profile(ctx):
    """
    Manage auth profiles.
    """
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())
        sys.exit(0)


def _load_all_on_disk_profiles() -> dict:
    candidate_profile_names = Profile.list_on_disk_profiles()
    profiles_dicts = OrderedDict()
    for candidate_profile_name in candidate_profile_names:
        try:
            # conf = Profile.load_auth_client_config(candidate_profile_name)
            _, conf = PlanetAuthFactory.load_auth_client_config_from_profile(candidate_profile_name)
            profiles_dicts[candidate_profile_name] = conf
        except Exception as ex:
            auth_logger.debug(
                msg=f'"{candidate_profile_name}" was not a valid locally defined profile directory: {ex}'
            )

    return profiles_dicts


@cmd_profile.command("list")
@opt_long
@recast_exceptions_to_click(AuthException, FileNotFoundError, PermissionError)
def cmd_profile_list(long):
    """
    List auth profiles.
    """
    if long:
        click.echo("Built-in profiles:")
        profile_names = Builtins.builtin_profile_names().copy()
        profile_names.sort()
        display_object = OrderedDict()
        for profile_name in profile_names:
            config_dict = Builtins.builtin_profile_auth_client_config_dict(profile_name)
            display_object[profile_name] = config_dict
        print_obj(display_object)

        click.echo("\nLocally defined profiles:")
        print_obj(_load_all_on_disk_profiles())

    else:
        click.echo("Built-in profiles:")
        profile_names = Builtins.builtin_profile_names().copy()
        profile_names.sort()
        print_obj(profile_names)
        click.echo("\nLocally defined profiles:")
        profile_names = Profile.list_on_disk_profiles()
        profile_names.sort()
        print_obj(profile_names)


@cmd_profile.command("create")
@opt_sops
@click.argument("new_profile_name", required=False)
@recast_exceptions_to_click(AuthException, FileNotFoundError, PermissionError)
def cmd_profile_create(sops, new_profile_name):
    """
    Wizard to create a new authentication profile.
    """
    # TODO: Non-interactive form?  The required args would be different for each type.
    #       If we conditionally prompt, what is the story? --no-sops? --quiet?  Pre-populate promts from CLI?
    # TODO: prompt for sops?
    # TODO: unify headless and interactive CL arg handling.
    # TODO: have a profile edit command that primes from an existing profile
    # TODO: Should the defaults not be meta but a more core feature
    # TODO: wrap text in narrow TTYs
    # TODO: rename config_hints meta to wizard hints?  This is NOT a config schema
    # TODO: can we take the config hint defaults from a populated dictionary? (meta['config_defaults'] ? )
    # TODO: We have no handling of non string types (Notably, handling "scopes" would be nice)

    if not new_profile_name:
        new_profile_name = _dialogue_enter_auth_profile_name()

    config_type, client_type = _dialogue_choose_auth_client_type()  # pylint: disable=W0612

    if sops:
        dst_config_filepath = Profile.get_profile_file_path(profile=new_profile_name, filename=AUTH_CONFIG_FILE_SOPS)
    else:
        dst_config_filepath = Profile.get_profile_file_path(profile=new_profile_name, filename=AUTH_CONFIG_FILE_PLAIN)

    config_dict = {
        AuthClientConfig.CLIENT_TYPE_KEY: config_type.meta().get("client_type"),
    }

    if config_type.meta().get("config_hints"):
        for hint in config_type.meta().get("config_hints"):
            config_value = input_dialog(
                title="{} Configuration: {}".format(config_type.meta().get("display_name"), hint.get("config_key")),
                text="{} ({})\n{}".format(
                    hint.get("config_key_name"), hint.get("config_key"), hint.get("config_key_description")
                ),
                default=hint.get("config_key_default") or "",
                cancel_text="Skip",  # since we keep going...
            ).run()  # or _handle_canceled() Users can skip config values.
            config_dict[hint.get("config_key")] = config_value

    new_auth_client_config = config_type(file_path=dst_config_filepath, **config_dict)
    new_auth_client_config.save()


@cmd_profile.command("edit")
@recast_exceptions_to_click(AuthException, FileNotFoundError, PermissionError)
def cmd_profile_edit():
    """
    Edit an existing profile.
    """
    raise AuthException("Function not implemented")


@cmd_profile.command("copy")
@click.argument("src")
@click.argument("dst")
@opt_sops
@recast_exceptions_to_click(AuthException, FileNotFoundError, PermissionError)
def cmd_profile_copy(sops, src, dst):
    """
    Copy an existing profile to create a new profile.  Only the persistent
    profile configuration will be copied.  User access tokens initialized
    via a call to `login` will not be copied.  Note: Depending on the
    type of [planet_auth.AuthClient] configured in the source profile,
    the new profile may have long term credentials (e.g. OAuth
    client credential secrets, API keys. etc.).  External support files,
    such as public/private keypair files, are not copied.

    This command will work with built-in as well as custom profiles,
    so it is possible to bootstrap profiles to manage multiple user
    identities with an otherwise default client profile:
    ```
    profile copy default <my_new_profile>
    profile copy legacy  <my_new_profile>
    ```

    """
    # TODO: consider fixing the copying of support files like key pairs when
    #       pubkeys are used.  To do that properly, their paths should be
    #       relative to the profile dir, and not absolute, but that is not
    #       currently implemented.
    _, auth_config = PlanetAuthFactory.load_auth_client_config_from_profile(src)
    if sops:
        dst_config_filepath = Profile.get_profile_file_path(profile=dst, filename=AUTH_CONFIG_FILE_SOPS)
    else:
        dst_config_filepath = Profile.get_profile_file_path(profile=dst, filename=AUTH_CONFIG_FILE_PLAIN)

    auth_config.set_path(dst_config_filepath)
    # No need to specify provider. The CLI is only concerned with the
    # default file based storage provider for now.
    # auth_config.auth_client().config().set_storage_provider()
    auth_config.save()


@cmd_profile.command("set")
@click.argument("selected_profile", required=False)
@recast_exceptions_to_click(AuthException, FileNotFoundError, PermissionError)
def cmd_profile_set(selected_profile):
    """
    Configure the default authentication profile to use when one is not otherwise specified.
    """
    if not selected_profile:
        selected_profile = _dialogue_choose_auth_profile()

    try:
        user_profile_config_file = PlanetAuthUserConfig()
        user_profile_config_file.load()
    except FileNotFoundError:
        user_profile_config_file = PlanetAuthUserConfig(data={})

    user_profile_config_file.update_data({EnvironmentVariables.AUTH_PROFILE: selected_profile})
    user_profile_config_file.save()


@cmd_profile.command("show")
@click.pass_context
@recast_exceptions_to_click(AuthException, FileNotFoundError, PermissionError)
def cmd_profile_show(ctx):
    """
    Show the current authentication profiles.
    """
    print(f'Current: {ctx.obj["AUTH"].profile_name()}')
    try:
        user_profile_config_file = PlanetAuthUserConfig()
        print(f"User Default: {user_profile_config_file.lazy_get(EnvironmentVariables.AUTH_PROFILE)}")
    except Exception:  #  as ex:
        # print(f'User Default: {ex}')
        print("User Default: N/A")

    print(f"Global Built-in Default: {Builtins.builtin_default_profile_name()}")



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jwt
from typing import Dict, List, Optional, Tuple

import planet_auth.logging.auth_logger
from planet_auth import ExpiredTokenException, TokenValidatorException, InvalidArgumentException
from planet_auth.auth import Auth
from planet_auth.auth_client import AuthClient
from planet_auth.auth_exception import AuthException, InvalidTokenException
from planet_auth.logging.events import AuthEvent
from planet_auth.oidc.api_clients.introspect_api_client import IntrospectionApiClient
from planet_auth.oidc.auth_client import OidcAuthClient

auth_logger = planet_auth.logging.auth_logger.getAuthLogger()


class OidcMultiIssuerValidator:
    """
    The OidcMultiIssuerValidator is a token validator that can be configured
    to trust multiple token issuers. This is not expected to be a normal
    operating mode for most services. This was developed to support migration
    use cases.

    This is a higher level utility class, and is built on top of
    [planet_auth.AuthClient][] classes.  For a lower level utilities,
    see [planet_auth.TokenValidator][], or see the validation
    functionality built into [planet_auth.AuthClient][].

    Warning:
        Never ever ever EVER use this to bridge between trust environments.
        For example, do not use this to trust staging and production environment
        issuing authorities.
    """

    # TODO: Should we consider taking AuthClient rather than Auth type?
    #       We only ever use the client (but Auth is the more friendly
    #       higher level aggregation)
    #       The use case of validating incoming request doesn't really
    #       require the the outbound request authentication capabilities
    #       provided by the Auth type.
    # TODO: The lower level TokenValidtor allows the caller to specify
    #       what token signing algorithms are trusted.  This class does
    #       not. It's not exactly straight forward what changes to make
    #       since we don't use the Token Validator directly.  Rather,
    #       we work through the wrapper Auth and AuthClient types so
    #       that we have access to remote validators and possibly could
    #       handle opaque tokens where "algorithm" doesn't really apply.
    def __init__(
        self,
        trusted: List[Auth],
        log_result: bool = True,
    ):
        """
        Create a new multi issuer validator using the provided auth clients.
        The auth clients are expected to be any Auth that implements OIDC functionality.

        Since the expected audience could be different for each issuer,
        the expectation is that the provided Auth contexts will be configured
        with an audience, even though that is an optional configuration parameter
        for most implementing classes.

        Parameters:
            trusted:
            log_result:
        """
        self._trusted: Dict[str, Auth] = {}
        self._log_result = log_result

        def _check_auth_client(auth_client: AuthClient) -> OidcAuthClient:
            # pylint: disable=protected-access
            if not isinstance(auth_client, OidcAuthClient):
                raise AuthException(message="Auth Provider must be an OIDC auth provider type")
            if not auth_client._oidc_client_config.audiences():
                raise AuthException(
                    message="Auth Providers used for OIDC token validation must have the audiences configuration value set."
                )
            return auth_client

        for auth_provider in trusted:
            if auth_provider:
                auth_client = _check_auth_client(auth_provider.auth_client())
                issuer = auth_client._issuer()
                if issuer in self._trusted:
                    raise AuthException(
                        message="Cannot configure multiple auth providers for the same issuer '{}'".format(issuer)
                    )
                self._trusted[issuer] = auth_provider

    # TODO: we should probably deprecate this method...
    @staticmethod
    def from_auth_server_urls(
        trusted_auth_server_urls: List[str],
        audience: str = None,
        log_result: bool = True,
    ):
        """
        Create a new multi issuer validator from the provided OAuth server URLs.
        This is a convenience method for common cases.

        Warning:
            When mutli validators are initialized from URLs by this method, it is
            assumed that the auth server URL (used to perform OIDC discovery and
            locate all the other API endpoints) and the issuer (which is burned
            into the signed access tokens) are the same. This is normally
            true, but not universally required. For example, some network
            proxy configurations may make these different.

            This assumption is made in part to avoid the discovery lookup
            during construction. Since a network lookup could potentially fail,
            construction of the entire configuration could be rendered non-functional
            by a network or configuration problem impacting a single auth server.

            The multi-validator requires foreknowledge of all the issuers for
            proper initialization. So, without this assumption it would be unavoidable
            to introduce network risk into the constructor.  This assumption
            allows us to push all network errors to runtime, and avoids
            possible initialization time errors.
        """
        trusted = []
        for auth_server_url in trusted_auth_server_urls:
            if auth_server_url:
                trusted.append(
                    Auth.initialize_from_config_dict(
                        client_config={
                            "client_type": "oidc_client_validator",
                            "auth_server": auth_server_url,
                            "issuer": auth_server_url,
                            "audiences": [audience],
                        }
                    )
                )

        return OidcMultiIssuerValidator(
            trusted=trusted,
            log_result=log_result,
        )

    @staticmethod
    def from_auth_server_configs(
        trusted_auth_server_configs: List[dict],
        log_result: bool = True,
    ):
        """
        Create a new multi issuer validator from the provided auth client
        config dictionaries.

        Parameters:
            trusted_auth_server_configs: A list of configuration dictionaries.
                Unless remote validation is required, the configuration dictionaries
                may be sparse, containing only the `auth_server` and `audiences` properties.
                `auth_server` is expected to be a single string, containing the URL
                of the OAuth issuer.  `audiences` is expected to be an array, and contain
                a list of supported audiences.
            log_result: Control whether successful token validations against
                trusted auth servers should be logged.

        Example:
            ```python
            auth_validator = planet_auth.OidcMultiIssuerValidator.from_auth_server_configs(
                trusted_auth_server_configs=[
                    {
                        "auth_server": "https://account-next.planet.com/oauth2/auth_server_id",
                        "audiences": ["https://api.staging.planet-labs.com/"],
                    },
                ],
            )
            ```
        """
        trusted = []

        def _doctor_config(auth_server_conf):
            # patch up sparse configs.
            if not auth_server_conf.get("client_type"):
                auth_server_conf["client_type"] = "oidc_client_validator"
            if not auth_server_conf.get("issuer"):
                # Optimization. Might not be universally safe.  Short circuits
                # discovery and has all the implications discussed above.
                auth_server_conf["issuer"] = auth_server_conf.get("auth_server")

        for auth_server_conf in trusted_auth_server_configs:
            if auth_server_conf:
                _doctor_config(auth_server_conf)
                trusted.append(Auth.initialize_from_config_dict(client_config=auth_server_conf))

        return OidcMultiIssuerValidator(
            trusted=trusted,
            log_result=log_result,
        )

    @staticmethod
    def _check_access_token(
        token: str, auth_client: AuthClient, do_remote: bool, scopes_anyof: list = None
    ) -> Tuple[Dict, Optional[Dict]]:
        # The client MUST be configured with the expected audience for local validation
        local_validation = auth_client.validate_access_token_local(access_token=token, scopes_anyof=scopes_anyof)
        if not local_validation:
            # We expect the auth client to throw when the token cannot be validated.
            # Protect ourselves against misbehaving implementations.
            raise AuthException(message="Internal error. No claims could be validated, but no error was reported.")

        remote_validation = None
        if do_remote:
            remote_validation = auth_client.validate_access_token_remote(access_token=token)
            # This will raise on invalid results.
            # The OIDC implementations should already do this, but we double-check
            # so that MultiIssuerValidator can work with arbitrary
            # auth clients whose implementations we may not know.
            IntrospectionApiClient.check_introspection_response(remote_validation)

        return local_validation, remote_validation

    def _select_validator(self, token) -> Auth:
        # WARNING: Treat unverified token claims like toxic waste.
        #          Nothing can be trusted until the token is verified.
        unverified_decoded_token = jwt.decode(token, options={"verify_signature": False})  # nosemgrep
        issuer = unverified_decoded_token.get("iss")
        if not issuer:
            # PyJWT does not seem to raise if the issuer is explicitly None, even when
            # verify_iss was selected.
            raise InvalidTokenException(message="Cannot validate token that does not include an issuer ('iss') claim")
        if not isinstance(issuer, str):
            raise InvalidTokenException(
                message=f"Issuer claim ('iss') must be a of string type. '{type(issuer).__name__}' type was detected."
            )

        validator = self._trusted.get(issuer)
        if validator:
            return validator
        raise AuthException(
            message="Rejecting token from an unrecognized issuer '{}'".format(issuer),
            event=AuthEvent.TOKEN_INVALID_BAD_ISSUER,
        )

    # TODO?: remote revocation should cache failures to protect auth
    #        from abuse.  We never expect invalid tokens to become valid.
    #        (Should this be a concern of the inner validators, or this higher class?)
    # TODO: add required_claims_TODO or the like?  Need to nail down
    #       the exact contract this lib provides.  AuthN vs AuthZ.
    #       As of now, we are drawing the line as scope claims.
    @auth_logger.log_exception(default_event=AuthEvent.TOKEN_INVALID, exception_cls=InvalidTokenException)
    @TokenValidatorException.recast(jwt.PyJWTError)
    @InvalidTokenException.recast(jwt.InvalidTokenError)
    @ExpiredTokenException.recast(jwt.ExpiredSignatureError)
    def validate_access_token(
        self, token: str, do_remote_revocation_check: bool = False, scopes_anyof: list = None
    ) -> Tuple[Dict, Optional[Dict]]:
        """
        Validate an access token, and return the token's claims if the
        all validation has succeeded.

        Remote revocation checks should be done for high value operations
        where it is undesirable to allow revoked tokens to be used
        for the remainder of their lifetime.  Token lifetimes
        should be chosen to be an acceptable trade-off between
        permitting revoked tokens for the remainder of the token lifespan
        for performance and scalability, and performing remote validation
        for extra security.

        When an invalid token is presented, an exception is thrown.

        Parameters:
            token: Token to validate
            do_remote_revocation_check: Control whether the issuer will be
                consulted for revocation of the access token.
            scopes_anyof: Optional list of OAuth2 scopes to check for in the token.
                This list is an "any of" list of scopes. As long as one of the scopes in the
                list is present, validation will pass. If none of the scopes are present,
                validation will fail.
        Returns:
            Returns a tuple of validation results, as returned from
                [planet_auth.AuthClient.validate_access_token_local][]
                and [planet_auth.AuthClient.validate_access_token_remote][]
        """

        if not token:
            raise InvalidArgumentException(message="Cannot validate empty string as a token")

        validator = self._select_validator(token)
        local_validation, remote_validation = self._check_access_token(
            token=token,
            auth_client=validator.auth_client(),
            do_remote=do_remote_revocation_check,
            scopes_anyof=scopes_anyof,
        )

        if self._log_result:
            auth_logger.info(
                msg="Accepting access token from a primary issuing authority",
                event=AuthEvent.TOKEN_VALID,
                # FIXME? - We do not currently pass this info up from the validators.
                # jwt_header_json={"alg": "RS256"},
                jwt_body_json=local_validation,
            )

        return local_validation, remote_validation



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from planet_auth.credential import Credential
from planet_auth.storage_utils import InvalidDataException, ObjectStorageProvider


class FileBackedPlanetLegacyApiKey(Credential):
    """
    Credential object for storing Planet Legacy API Keys
    """

    def __init__(
        self, api_key=None, jwt=None, api_key_file=None, storage_provider: Optional[ObjectStorageProvider] = None
    ):
        if api_key or jwt:
            # "key" was used by the old Python SDK.
            # Keeping this json key name and file format so this class
            # should just work with tokens saved by pre-OAuth versions of the SDK.
            init_data = {"key": api_key, "jwt": jwt}
        else:
            init_data = None
        super().__init__(data=init_data, file_path=api_key_file, storage_provider=storage_provider)

    def check_data(self, data):
        """
        Check that the supplied data represents a valid Planet Legacy API Key object.
        """
        super().check_data(data)
        if not data.get("key"):
            raise InvalidDataException(message="'key' not found in file " + str(self._file_path))

    def legacy_api_key(self):
        """
        Get the current API key.
        """
        return self.lazy_get("key")

    # Duck type compatibility with FileBackedApiKey (Static API key)
    def api_key(self):
        return self.legacy_api_key()

    def legacy_jwt(self):
        """
        Get the saved legacy JWT.
        """
        return self.lazy_get("jwt")



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import getpass
import pathlib
import jwt
from requests import Session
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from typing import Optional, Union

import planet_auth.logging.auth_logger
from planet_auth.auth_client import AuthClientConfig, AuthClient, AuthClientException, AuthClientConfigException
from planet_auth.constants import X_PLANET_APP_HEADER, X_PLANET_APP
from planet_auth.credential import Credential
from planet_auth.planet_legacy.legacy_api_key import FileBackedPlanetLegacyApiKey
from planet_auth.planet_legacy.request_authenticator import PlanetLegacyRequestAuthenticator
from planet_auth.request_authenticator import CredentialRequestAuthenticator
from planet_auth.util import parse_content_type

auth_logger = planet_auth.logging.auth_logger.getAuthLogger()


class PlanetLegacyAuthClientConfig(AuthClientConfig):
    """
    Configuration required for [planet_auth.PlanetLegacyAuthClient][]
    """

    def __init__(self, legacy_auth_endpoint=None, api_key=None, **kwargs):
        super().__init__(**kwargs)
        if api_key:
            self._data["api_key"] = api_key
        if legacy_auth_endpoint:
            self._data["legacy_auth_endpoint"] = legacy_auth_endpoint

    def check_data(self, data):
        super().check_data(data)
        if not data.get("legacy_auth_endpoint"):
            raise AuthClientConfigException(
                message="legacy_auth_endpoint must be configured for Planet Legacy auth clients."
            )

    def api_key(self) -> str:
        return self.lazy_get("api_key")

    def legacy_auth_endpoint(self) -> str:
        return self.lazy_get("legacy_auth_endpoint")

    @classmethod
    def meta(cls):
        return {
            "client_type": "planet_legacy",
            "auth_client_class": PlanetLegacyAuthClient,
            "display_name": "Planet Legacy Auth",
            "description": "Planet legacy authentication protocols.  Not recommended for new applications.",
            "config_hints": [
                {
                    "config_key": "legacy_auth_endpoint",
                    "config_key_name": "Planet Legacy Authentication Endpoint",
                    "config_key_description": "API endpoint used to perform user authentication against the Planet cloud service.",
                    # "config_key_default": "",
                },
                {
                    "config_key": "api_key",
                    "config_key_name": "Planet API Key",
                    "config_key_description": "Planet API key used to authenticate requests",
                },
            ],
        }


class PlanetLegacyAuthClientException(AuthClientException):
    def __init__(self, raw_response=None, **kwargs):
        super().__init__(**kwargs)
        self.raw_response = raw_response


class PlanetLegacyAuthClient(AuthClient):
    """
    Implementation of the AuthClient that interacts with Planet's API key
    auth interfaces.
    """

    def __init__(self, legacy_client_config: PlanetLegacyAuthClientConfig):
        super().__init__(legacy_client_config)
        self._legacy_client_config = legacy_client_config

        retry_strategy = Retry(total=3, backoff_factor=1, status_forcelist=[429], allowed_methods=["POST", "GET"])
        adapter = HTTPAdapter(max_retries=retry_strategy)
        self._session = Session()
        self._session.mount("https://", adapter)
        # self._session.mount("http://", adapter)

    @staticmethod
    def _prepare_auth_payload(username, password):
        data = {"email": username, "password": password}
        return data

    def _check_http_error(self, response):
        if not response.ok:
            raise PlanetLegacyAuthClientException(
                message="HTTP error from endpoint at {}: {}: {}".format(
                    self._legacy_client_config.legacy_auth_endpoint(), response.status_code, response.reason
                ),
                raw_response=response,
            )

    def _check_json_payload(self, response):
        json_response = None
        if response.content:
            ct = parse_content_type(response.headers.get("content-type"))
            if not ct["content-type"] == "application/json":
                raise PlanetLegacyAuthClientException(
                    message='Expected json content-type, but got "{}"'.format(response.headers.get("content-type")),
                    raw_response=response,
                )
            json_response = response.json()
        if not json_response:
            raise PlanetLegacyAuthClientException(
                message="Response from authentication endpoint {}"
                " was not understood. Expected JSON response payload,"
                " but none was found.".format(self._legacy_client_config.legacy_auth_endpoint()),
                raw_response=response,
            )
        return json_response

    @staticmethod
    def _parse_json_response(response, json_response):
        token_jwt = json_response.get("token")
        if not token_jwt:
            raise PlanetLegacyAuthClientException(
                message='Authorization response did not include expected field "token"', raw_response=response
            )

        # The token is signed with a symmetric key.  The client does not
        # possess this key, and cannot verify the JWT.
        decoded_jwt = jwt.decode(token_jwt, options={"verify_signature": False})  # nosemgrep
        api_key = decoded_jwt.get("api_key")
        if not api_key:
            raise PlanetLegacyAuthClientException(
                message='Authorization response did not include expected field "api_key" in the returned token',
                raw_response=response,
            )

        return api_key, token_jwt

    def _checked_auth_request(self, auth_data):
        # Optimized for client use. We don't use a Session to pool connection
        # use at this time.
        response = self._session.post(
            self._legacy_client_config.legacy_auth_endpoint(),
            json=auth_data,
            headers={X_PLANET_APP_HEADER: X_PLANET_APP},
        )
        self._check_http_error(response)
        json_response = self._check_json_payload(response)
        return self._parse_json_response(response, json_response)

    def login(
        self,
        allow_open_browser: Optional[bool] = False,
        allow_tty_prompt: Optional[bool] = False,
        username: str = None,
        password: str = None,
        **kwargs,
    ) -> FileBackedPlanetLegacyApiKey:
        """
        Perform a login using Planet Legacy authentication endpoints.
        Parameters:
            username: Planet account user name. If not specified,
                the user will be prompted for their user name.
            password: Planet user password.  If not specified, the user
                will be prompted for their password.
            allow_tty_prompt: specify whether login is permitted to request
                input from the terminal.
        Returns:
            Upon successful login, a Credential will be returned. The returned
                value will be in memory only. It is the responsibility of the
                application to save this credential to disk as appropriate using
                the mechanisms built into the Credential type.
        """
        # TODO: should we warn if someone is doing a login when a static key has been
        #       configured in the client config (as opposed to having been saved
        #       in a credential file as the result of a login)?
        if not username:
            if allow_tty_prompt:
                username = input("Email: ")
            else:
                raise PlanetLegacyAuthClientException(
                    message="Username must be provided when performing non-interactive login"
                )

        if not password:
            if allow_tty_prompt:
                password = getpass.getpass(prompt="Password: ")
            else:
                raise PlanetLegacyAuthClientException(
                    message="Password must be provided when performing non-interactive login"
                )

        auth_payload = self._prepare_auth_payload(username, password)
        api_key, returned_jwt = self._checked_auth_request(auth_payload)
        return FileBackedPlanetLegacyApiKey(api_key=api_key, jwt=returned_jwt)

    def default_request_authenticator(
        self, credential: Union[pathlib.Path, Credential]
    ) -> CredentialRequestAuthenticator:
        # If an API key has been configured in the client config, use that and ignore
        # any separate credential file.
        storage_provider = self._auth_client_config.storage_provider()
        if self._legacy_client_config.api_key():
            _credential = FileBackedPlanetLegacyApiKey(
                api_key=self._legacy_client_config.api_key(),
                api_key_file=self._legacy_client_config.path(),
                storage_provider=storage_provider,
            )
        else:
            if isinstance(credential, pathlib.Path):
                _credential = FileBackedPlanetLegacyApiKey(api_key_file=credential, storage_provider=storage_provider)
            elif isinstance(credential, FileBackedPlanetLegacyApiKey):
                _credential = credential
            elif credential is None:
                # This will be brain-dead until update_credential() or update_credential_data()
                # is called.  This is useful for initializing properly typed credential objects.
                _credential = FileBackedPlanetLegacyApiKey(storage_provider=storage_provider)
            else:
                raise TypeError(
                    f"{type(self).__name__} does not support {type(credential)} credentials.  Use file path or FileBackedPlanetLegacyApiKey."
                )

        return PlanetLegacyRequestAuthenticator(planet_legacy_credential=_credential)

    def can_login_unattended(self) -> bool:
        # We could allow username/password to be specified in the client
        # config, and forgo prompts in login(), but we've not implemented
        # what would be a bad security practice of saving passwords.
        return False



# setup.py is currently required for installing in editable mode ("pip install -e .").
# Otherwise, most of this project has had its config moved to the more declarative
# pyproject.toml

from setuptools import setup

setup()



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations  # https://stackoverflow.com/a/33533514

from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Union
import pathlib

import planet_auth.logging.auth_logger
from planet_auth.auth_exception import AuthException
from planet_auth.credential import Credential
from planet_auth.request_authenticator import CredentialRequestAuthenticator
from planet_auth.storage_utils import InvalidDataException, ObjectStorageProvider, FileBackedJsonObject

auth_logger = planet_auth.logging.auth_logger.getAuthLogger()


class AuthClientException(AuthException):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class AuthClientConfigException(InvalidDataException):  # Audit
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class AuthClientConfig(FileBackedJsonObject, ABC):
    """
    Base class for auth client configuration objects. Each concrete auth
    client type has a dedicated auth client config type to provide
    client specific validation and guardrails.

    The factory methods in the base class accept a dictionary, and will
    return an instance of an appropriate subclass.
    """

    _typename_map: Dict[str, AuthClientConfig] = {}
    _config_types: List[AuthClientConfig] = []
    # TODO: Constants for other keys? Most config dictionary keys belong to
    #       child classes, not us. We would also have to distinguish meta keys
    #       from config dictionary keys. Most meta keys are not for the
    #       AuthClient itself, but helpers that may exist in other class
    #       hierarchies.
    CLIENT_TYPE_KEY = "client_type"

    def __init__(self, file_path=None, storage_provider: Optional[ObjectStorageProvider] = None, **kwargs):
        super().__init__(data=kwargs, file_path=file_path, storage_provider=storage_provider)
        if len(kwargs.keys()) > 0:
            # raise AuthClientConfigException(
            # message='Unexpected config arguments in client configuration: {}'
            # .format(', '.join(kwargs.keys())))
            for key in kwargs.keys():
                if key != AuthClientConfig.CLIENT_TYPE_KEY:
                    auth_logger.debug(msg='Ignoring unknown keyword argument: "{}"'.format(str(key)))
                    # TODO? We now have the option of shoving these into self._data
                    # This will impact how we save/copy data

    @classmethod
    def _get_config_types(cls):
        if not cls._config_types:
            from planet_auth.oidc.auth_clients.client_validator import OidcClientValidatorAuthClientConfig
            from planet_auth.oidc.auth_clients.auth_code_flow import (
                AuthCodeClientConfig,
                AuthCodeWithClientSecretClientConfig,
                AuthCodeWithPubKeyClientConfig,
            )
            from planet_auth.oidc.auth_clients.client_credentials_flow import (
                ClientCredentialsPubKeyClientConfig,
                ClientCredentialsClientSecretClientConfig,
            )
            from planet_auth.oidc.auth_clients.device_code_flow import (
                DeviceCodeClientConfig,
                DeviceCodeWithClientSecretClientConfig,
                DeviceCodeWithPubKeyClientConfig,
            )
            from planet_auth.oidc.auth_clients.resource_owner_flow import (
                ResourceOwnerClientConfig,
                ResourceOwnerWithClientSecretClientConfig,
                ResourceOwnerWithPubKeyClientConfig,
            )
            from planet_auth.planet_legacy.auth_client import PlanetLegacyAuthClientConfig
            from planet_auth.static_api_key.auth_client import StaticApiKeyAuthClientConfig
            from planet_auth.none.noop_auth import NoOpAuthClientConfig

            cls._config_types = [
                AuthCodeClientConfig,
                AuthCodeWithPubKeyClientConfig,
                AuthCodeWithClientSecretClientConfig,
                ClientCredentialsPubKeyClientConfig,
                ClientCredentialsClientSecretClientConfig,
                DeviceCodeClientConfig,
                DeviceCodeWithClientSecretClientConfig,
                DeviceCodeWithPubKeyClientConfig,
                ResourceOwnerClientConfig,
                ResourceOwnerWithPubKeyClientConfig,
                ResourceOwnerWithClientSecretClientConfig,
                OidcClientValidatorAuthClientConfig,
                PlanetLegacyAuthClientConfig,
                StaticApiKeyAuthClientConfig,
                NoOpAuthClientConfig,
            ]
        return cls._config_types

    @classmethod
    def _get_typename_map(cls):
        if not cls._typename_map:
            for conf_type in cls._get_config_types():
                cls._typename_map[conf_type.meta().get(cls.CLIENT_TYPE_KEY)] = conf_type
        return cls._typename_map

    @classmethod
    def from_dict(cls, config_data: Dict) -> AuthClientConfig:
        """
        Create a AuthClientConfig from a configuration dictionary.
        Returns:
            A concrete auth client config object.
        """
        if not config_data:
            raise AuthClientException(message="Error: Auth client config dictionary must not be empty or None")
        config_type = config_data.get(cls.CLIENT_TYPE_KEY)
        config_cls = AuthClientConfig._get_typename_map().get(config_type)
        if not config_cls:
            raise AuthClientException(
                message='Error: Auth client config type "{}" is not understood by the factory.'.format(config_type)
            )
        return config_cls(**config_data)

    @staticmethod
    def from_file(file_path, storage_provider: Optional[ObjectStorageProvider] = None) -> AuthClientConfig:
        """
        Create an AuthClientConfig from a json file that contains a config
        dictionary.
        Returns:
            A concrete auth client config object.
        """
        config_file = FileBackedJsonObject(file_path=file_path, storage_provider=storage_provider)
        config_file.load()
        # AuthClientConfig.from_dict() also returns a FileBackedJsonObject.
        # Re-creating with from_dict() after the load() gives us a more strongly
        # typed subclass with better data checking and better error messages.
        conf = AuthClientConfig.from_dict(config_file.data())
        conf.set_path(file_path)
        conf.set_storage_provider(storage_provider=storage_provider)
        return conf

    @classmethod
    @abstractmethod
    def meta(cls) -> Dict:
        """
        Return a dictionary of metadata.
        The meta dictionary provides a place to store information that is
        primarily for users of AuthClient types rather than for the operation
        of the Auth Client itself.

        Meta Keys:
            client_type: The literal value used for `client_type` in configuration
                dictionaries that indicates the selection of the AuthClient type.

            auth_client_class: The concrete `AuthClient` class that is served by
                the configuration type.

            display_name: Display name for the `AuthClient` type

            description: A description of the `AuthClient` type

            config_hints: A list configuration hints bootstrapping a configuration
                with a wizard of some sort.  This is distinct from a rigid schema for
                the configuration type.
        """


# TODO: Revisit this interface.  Since we've done some other refactoring,
#       it may not make sense for all of these to be on the AuthClient
#       base ABC. (Specifically, many of the OAuth specific methods,
#       perhaps others.)
class AuthClient(ABC):
    """
    Base class for auth clients.  Concrate instances of this base class
    manage the specific of how to authenticate a user and obtain credentials
    that may be used for service APIs.

    The factory methods in the base class accepts a client specific client
    configuration type, and will return an instance of an appropriate subclass.

    Example:
        ```python
        from planet_auth import AuthClientConfig, AuthClient
        config_dict = { ... } # See AuthClientConfig
        my_auth_config = AuthClientConfig.from_dict(config_dict)
        my_auth_client = AuthClient.from_config(my_auth_config)
        ```
    """

    _type_map: Dict[AuthClientConfig, AuthClient] = {}

    def __init__(self, auth_client_config: AuthClientConfig):
        auth_client_config.lazy_load()
        auth_client_config.check()  # Cowardly refuse to create clients with brain damaged configuration.
        self._auth_client_config = auth_client_config

    @classmethod
    def _get_type_map(cls):
        # pylint: disable=protected-access
        if not cls._type_map:
            for conf_type in AuthClientConfig._get_config_types():
                cls._type_map[conf_type] = conf_type.meta().get("auth_client_class")
        return cls._type_map

    @classmethod
    def from_config(cls, config: AuthClientConfig) -> AuthClient:
        """
        Create an AuthClient of an appropriate subtype from the client config.
        Returns:
            An initialized auth client instance.
        """
        client_cls = AuthClient._get_type_map().get(type(config))
        if not client_cls:
            raise AuthClientException(message="Error: Auth client config class is not understood by the factory.")

        return client_cls(config)

    def config(self) -> AuthClientConfig:
        return self._auth_client_config

    @abstractmethod
    def login(
        self, allow_open_browser: Optional[bool] = False, allow_tty_prompt: Optional[bool] = False, **kwargs
    ) -> Credential:
        """
        Perform an initial login using the authentication mechanism
        implemented by the AuthClient instance.  The results of a successful
        login is FileBackedJsonObject Credential containing credentials that
        may be used for subsequent service API requests.  How these
        credentials are used for this purpose is outside the scope of either
        the AuthClient or the Credential.  This is the job of a
        RequestAuthenticator.

        The login command is permitted to be user interactive.  Depending on
        the implementation, this may include terminal prompts, or may require the
        use of a web browser.

        Authentication parameters are specific to each implementation. Consult
        subclass documentation for details.

        Args:
            allow_open_browser: specify whether login is permitted to open
                a browser window.
            allow_tty_prompt: specify whether login is permitted to request
                input from the terminal.
        Returns:
            Upon successful login, a Credential will be returned. The returned
                value will be in memory only. It is the responsibility of the
                application to save this credential to disk as appropriate using
                the mechanisms built into the Credential type.  AuthClient
                implementations should raise an exception for all login errors.
        """

    def device_login_initiate(self, **kwargs) -> Dict:
        """
        Initiate the process to login a device with limited UI capabilities.
        The returned dictionary should contain information for the application
        to present to the user, allowing the user to complete the login process
        asynchronously.

        After prompting the user, `device_login_complete()` should be called
        with the same dictionary that was returned by this call.

        Returns:
            Upon successful initiation of an asynchronous device login process,
                a dictionary containing information that must be presented
                to the user will be returned.
        """
        raise AuthClientException(message="Device login is not supported for the current authentication mechanism")

    def device_login_complete(self, initiated_login_data: Dict) -> Credential:
        """
        Complete a login process that was initiated by a call to `device_login_initiate()`.

        Args:
            initiated_login_data: The dictionary that was returned by `device_login_initiate()`

        Returns:
            Upon successful login, a Credential will be returned. The returned
                value will be in memory only. It is the responsibility of the
                application to save this credential to disk as appropriate using
                the mechanisms built into the Credential type.
        """
        raise AuthClientException(message="Device login is not supported for the current authentication mechanism")

    def refresh(self, refresh_token: str, requested_scopes: List[str]) -> Credential:
        # TODO: It may be better to accept a Credential as input?
        """
        Obtain a refreshed credential using the supplied refresh token.
        This method will be implemented by concrete AuthClients that
        implement a particular OAuth flow.
        Parameters:
            refresh_token: Refresh token
            requested_scopes: Scopes to request in the access token
        Returns:
            Upon success, a fresh Credential will be returned. As with
                login(), this credential will not have been persisted to storage.
                This is the responsibility of the application.
                AuthClient implementations should raise an exception for all login
                errors.
        """
        raise AuthClientException(message="Refresh not implemented for the current authentication mechanism")

    def validate_access_token_remote(self, access_token: str) -> Dict:
        """
        Validate an access token with the authorization server.
        Parameters:
            access_token: Access token to validate
        Returns:
            Returns a dictionary of validated token claims. It should be
                noted that the returned dictionaries from local and service
                token validation are not exactly the same.  Service validation
                should return a RFC 7662 dictionary.  Local validation returns
                the token claims.
        """
        raise AuthClientException(
            message="Access token validation is not implemented for the current authentication mechanism"
        )

    def validate_access_token_local(
        self, access_token: str, required_audience: str = None, scopes_anyof: list = None
    ) -> Dict:
        """
        Validate an access token locally. While the validation is local,
        the authorization server may still may contacted to obtain signing
        keys for validation.  Signing keys will be cached for future use.

        Auth servers (issuers) may provide service for any number of audiences,
        so which audience is expected / required is potentially different
        for each discrete validation check.  If the required audience
        is not provided as an argument, the validate should fall back
        to the audiences configured in the AuthClientConfig.

        If the audience is neither provided as an argument nor present in
        the AuthClientConfig, an error should be raised.

        Note:
            While tokens may have multiple audiences, and AuthClients
            may be configured to request multiple audiences,
            the validation method currently only supports checking for
            a single audience. This could be considered is a bit of an
            API mismatch. However, at this time this is considered expected
            and desirable behavior. Validation is generally considered to
            be occurring in a context that self identifies as a particular
            audience, and overlapping claims may mean different things
            to different audiences.  Validation should be done in the
            context of a single audience at a time.

        Parameters:
            access_token: Access token to validate.
            required_audience: Audience that the token is required to have.
            scopes_anyof: Optional list of OAuth2 scopes to check for in the token.
                This list is an "any of" list of scopes. As long as one of the scopes in the
                list is present, validation will pass. If none of the scopes are present,
                validation will fail.
        Returns:
            Returns a dictionary of validated token claims. It should be
                noted that the returned dictionaries from local and service
                token validation are not exactly the same.  Service validation
                should return a RFC 7662 dictionary.  Local validation returns
                the token claims.
        """
        raise AuthClientException(
            message="Access token validation is not implemented for the current authentication mechanism"
        )

    def validate_id_token_remote(self, id_token: str) -> Dict:
        """
        Validate an ID token with the authorization server.
        Parameters:
            id_token: ID token to validate
        Returns:
            Returns a dictionary of validated token claims
        """
        raise AuthClientException(
            message="ID token validation is not implemented for the current authentication mechanism"
        )

    def validate_id_token_local(self, id_token: str) -> Dict:
        """
        Validate an ID token locally. The authorization server may still be
        called to obtain signing keys for validation.  Signing keys will be
        cached for future use.
        Parameters:
            id_token: ID token to validate
        Returns:
            Returns a dictionary of validated token claims
        """
        raise AuthClientException(
            message="ID token validation is not implemented for the current authentication mechanism"
        )

    def validate_refresh_token_remote(self, refresh_token: str) -> Dict:
        """
        Validate a refresh token with the authorization server.
        Parameters:
            refresh_token: Refresh token to validate
        Returns:
            Returns the response from the remove validation service.
        """
        raise AuthClientException(
            message="Refresh token validation is not implemented for the current authentication mechanism"
        )

    def revoke_access_token(self, access_token: str):
        """
        Revoke an access token with the authorization server.
        Parameters:
            access_token: Access token to revoke.
        """
        raise AuthClientException(
            message="Access token revocation is not implemented for the current authentication mechanism"
        )

    def revoke_refresh_token(self, refresh_token: str):
        """
        Revoke a refresh token with the authorization server.
        Parameters:
            refresh_token: Access token to revoke.
        """
        raise AuthClientException(
            message="Refresh token revocation is not implemented for the current authentication mechanism"
        )

    def userinfo_from_access_token(self, access_token: str) -> Dict:
        """
        Look up user information from the auth server using the access token.
        Parameters:
            access_token: User access token.
        """
        raise AuthClientException(
            message="User information lookup is not implemented for the current authentication mechanism"
        )

    def oidc_discovery(self) -> Dict:
        """
        Query the authorization server's OIDC discovery endpoint for server information.
        Returns:
            Returns the OIDC discovery dictionary.
        """
        raise AuthClientException(
            message="OIDC discovery is not implemented for the current authentication mechanism."
        )

    # def oauth_discovery(self) -> Dict:
    #     """
    #     Query the authorization server's OAuth2 discovery endpoint for server information.
    #     Returns:
    #         Returns the OAuth2 discovery dictionary.
    #     """
    #     raise AuthClientException(
    #         message="OAuth2 discovery is not implemented for the current authentication mechanism."
    #     )

    def get_scopes(self) -> List[str]:
        """
        Query the authorization server for a list of scopes.
        Returns:
            Returns a list of scopes that may be requested during a call
                to login or refresh
        """
        raise AuthClientException(
            message="Listing scopes is not implemented for the current authentication mechanism."
        )

    # TODO: as a practical matter, I don't think we ever use the "Credential"
    #   option, and it makes the code more clumsy.  We should consider removing it.
    @abstractmethod
    def default_request_authenticator(
        self,
        credential: Union[pathlib.Path, Credential],
    ) -> CredentialRequestAuthenticator:
        """
        Return an instance of the default request authenticator to use for
        the specific AuthClient type and configured to use the provided
        credential file for persistence.

        It's not automatic that the default is always the right choice.
        Some authenticators may initiate logins, which may be interactive
        or not depending on the specifics of the AuthClient configuration
        and implementation type. Whether or not interactivity can be
        tolerated depends on the specifics of the surrounding application.

        In the simplest cases, there really is no relationship between
        the AuthClient and the request authenticator (see static key
        implementations). This relationship emerges when the mechanisms
        of an AuthClient requires frequent refreshing of the Credential.
        In these cases, it is convenient for the CredentialRequestAuthenticator
        to also have an AuthClient that is capable of performing this
        refresh transparently as needed.

        AuthClient implementors should favor defaults that do not require
        user interaction after an initial Credential has been obtained from
        an initial call to login()

        Parameters:
            credential: A Credential object, or a Path to the credential
                file that can be used to create an appropriate credential
                object.
        Returns:
            Returns an instance of the default request authenticator class to
                use for Credentials of the type obtained by the AuthClient.
        """

    @abstractmethod
    def can_login_unattended(self) -> bool:
        """
        Check whether the client can perform an unattended login.

        Depending on the implementation, some clients may be able to perform
        unattended logins based on information the client config (e.g. static
        API key clients can do this).  These should return true.  Other clients
        will always require user interactive logins (e.g. a user interactive
        client using a browser and MFA to login).  These should return false.
        Some clients may be flexible and support either depending on the state
        of their configuration.  These should return a result based on their
        configuration.
        """



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional

from planet_auth.oidc.api_clients.api_client import (
    OidcApiClient,
    OidcApiClientException,
    EnricherFuncType,
    _RequestParamsType,
    _RequestAuthType,
)


class DeviceAuthorizationApiException(OidcApiClientException):
    def __init__(self, message=None, raw_response=None):
        super().__init__(message, raw_response)


class DeviceAuthorizationApiClient(OidcApiClient):
    """
    Low level network client for the "device_authorization_endpoint" OAuth2/OIDC
    network endpoint.  This endpoint is defined by RFC 8628. See https://www.rfc-editor.org/rfc/rfc8628

    All invalid responses or error responses will result in an exception.
    """

    def __init__(self, device_authorization_uri: str):
        super().__init__(endpoint_uri=device_authorization_uri)

    @staticmethod
    def _prep_device_code_request_payload(
        client_id,
        requested_scopes: Optional[List[str]],
        requested_audiences: Optional[List[str]],
        extra: Optional[Dict],
    ) -> Dict:
        if extra is None:
            extra = {}
        # "None" is pythonic, and does not mean anything to OAuth APIs.
        _extra = {k: v for k, v in extra.items() if v is not None}

        data = {
            **_extra,  # we do this first, because we want it to lose the race with explicit parameters. It is not intended for clobbering.
            # "client_id": client_id,  # The job of the enricher.
        }
        if requested_scopes:
            data["scope"] = " ".join(requested_scopes)
        if requested_audiences:
            data["audience"] = " ".join(requested_audiences)

        return data

    @staticmethod
    def _check_device_auth_response(json_response: Dict) -> Dict:
        # Protocol endpoint specific response checks
        if not json_response.get("device_code"):
            raise DeviceAuthorizationApiException(
                "Unrecognized response: 'device_code' field was not present in the response."
            )
        if not json_response.get("user_code"):
            raise DeviceAuthorizationApiException(
                "Unrecognized response: 'user_code' field was not present in the response."
            )
        if not json_response.get("verification_uri"):
            raise DeviceAuthorizationApiException(
                "Unrecognized response: 'verification_uri' field was not present in the response."
            )
        if not json_response.get("expires_in"):
            raise DeviceAuthorizationApiException(
                "Unrecognized response: 'expires_in' field was not present in the response."
            )
        # verification_uri_complete and interval are optional under the spec, so we don't force them to be present.
        return json_response

    def _checked_request_device_code_call(
        self, request_params: _RequestParamsType, request_auth: Optional[_RequestAuthType]
    ) -> Dict:
        json_response = self._checked_post_json_response(request_params, request_auth)
        return self._check_device_auth_response(json_response)

    def request_device_code(
        self,
        client_id: str,
        requested_scopes: Optional[List[str]],
        requested_audiences: Optional[List[str]],
        auth_enricher: Optional[EnricherFuncType],
        extra,
    ) -> Dict:
        request_params = self._prep_device_code_request_payload(
            client_id=client_id,
            requested_scopes=requested_scopes,
            requested_audiences=requested_audiences,
            extra=extra,
        )
        request_auth = None
        if auth_enricher:
            request_params, request_auth = auth_enricher(request_params, self._endpoint_uri)

        return self._checked_request_device_code_call(request_params, request_auth)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List
from planet_auth.oidc.api_clients.api_client import OidcApiClient, OidcApiClientException


class JwksApiException(OidcApiClientException):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class JwksApiClient(OidcApiClient):
    """
    Low level API client for the OAuth JWKS endpoint.
    See RFC 7517 - JSON Web Key (JWK) and
    RFC 8414 - OAuth 2.0 Authorization Server Metadata
    for more information.
    """

    def __init__(self, jwks_uri: str):
        """
        Create a JWKS endpoint client
        """
        super().__init__(endpoint_uri=jwks_uri)

    def _checked_fetch(self):
        return self._checked_get_json_response(None, None)

    def jwks(self) -> Dict:
        """
        Fetch metadata from the JWKS endpoint.

        Returns:
            The full jwks response from the endpoint. An exception will be
                raised when an error occurs.
        """
        return self._checked_fetch()

    def jwks_keys(self) -> List:
        """
        Fetch keys from the JWKS endpoint.

        Returns:
            Just the jwks key set from the endpoint. An exception will be
                raised when an error occurs.
        """
        jwks_response = self.jwks()
        jwks_keys = jwks_response.get("keys")
        if not jwks_keys:
            raise JwksApiException(message='JWKS endpoint response did not include "keys" data')
        return jwks_keys



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import json
import logging
import importlib.metadata

from contextlib import suppress
from typing import Dict

from .events import AuthEvent
from planet_auth.auth_exception import AuthException, InvalidTokenException

# from planet_auth.oidc.token_validator import InvalidTokenException

# _lib_global_py_logger = logging.getLogger(__name__)

_lib_global_py_logger = logging.getLogger("planet_auth")
_lib_global_do_structured_logging = False
_lib_global_nested_logging_key = None

# Some services use the `json_logging` module which expects
# additional logging parameters to be stored under the 'props' key. This is
# the default for now so we don't break those dependant services' logging.
DEFAULT_NESTED_KEY = "props"


# class AuthLogger(logging.Logger):  # TODO?: is this a good idea, or a better approach?
class AuthLogger:
    """
    Class that wraps the Python logger so that all logs emitted from
    this library may be logged with the same consistent JSON format.
    This is being done so that dashboards may be built using structured
    data over string reg-ex parsing.
    """

    def __init__(self):
        # self._py_logger = logging.getLogger(__name__)
        # self._py_logger = logger
        self._msg_prefix = "[planet-auth-python] "
        self._auth_libraries = self._get_auth_libraries()

    def _get_auth_libraries(self):
        libs = {"planet-auth": importlib.metadata.version("planet-auth")}
        for optional_lib in (
            "planet-auth-config",
            "planet-auth-django",
        ):
            libs[optional_lib] = "N/A"
            with suppress(importlib.metadata.PackageNotFoundError):
                libs[optional_lib] = importlib.metadata.version(optional_lib)
        return libs

    def _get_py_logger(self):
        # _py_logger is not simply a member so that users of our library can call
        # setPyLoggerForAuthLogger and the library will work as expected.
        # if self._py_logger:
        #     return self._py_logger
        # pylint: disable=global-variable-not-assigned
        global _lib_global_py_logger
        return _lib_global_py_logger

    # TODO: should log level be encapsulated by the AuthLogger class?
    def log(
        self,
        level: int,
        msg: str = "",
        event: AuthEvent = None,
        jwt_header_json: Dict = None,
        jwt_body_json: Dict = None,
        exception: Exception = None,
    ) -> None:
        _logger = self._get_py_logger()
        if not _logger:
            return

        if level < _logger.getEffectiveLevel():
            return

        if exception:
            _log_msg = msg or str(exception)

            if not event and isinstance(exception, AuthException):
                event = exception.event()

            # Note: This is a little hacky. The lib is designed to handle more than just JWTs and OAuth,
            #     but it is a very common use case and this makes for an ergonomic development experience,
            #     making it easy to have the raise pass context in the exception to a distant point
            #     in the code responsible for logging.
            if isinstance(exception, InvalidTokenException):
                if not jwt_header_json:
                    jwt_header_json = exception.jwt_header()
                if not jwt_body_json:
                    jwt_body_json = exception.jwt_body()
        else:
            _log_msg = msg

        _log_msg = self._msg_prefix + _log_msg

        if not event:
            event = AuthEvent.TRACE

        log_json = {
            # "msg": _log_msg,
            "event": str(event),
            "auth_libraries": self._auth_libraries,
        }

        # TODO: Is this actually right?  The library is more general than OAuth and JWTs, but this
        #       is a common need when we log, so we've done this in our logger.
        if jwt_header_json:
            log_json["jwt_header"] = {"alg": jwt_header_json.get("alg")}
        if jwt_body_json:
            log_json["jwt_payload"] = {
                "iss": jwt_body_json.get("iss"),  # Standard claim
                "cid": jwt_body_json.get("cid"),  # Standard claim
                # "sub": jwt_body_json.get("sub"),  # Standard claim (may contain PII in some implementations)
                "aud": jwt_body_json.get("aud"),  # Standard claim
                "scope": jwt_body_json.get("scope"),  # RFC 8693, 9068 claim used for scope
                "scp": jwt_body_json.get("scp"),  # Okta claim used for scope
                "pl_principal": jwt_body_json.get("pl_principal"),  # Planet claim
                "organization_id": jwt_body_json.get("organization_id"),  # Planet claim
            }

        if exception:
            log_json["error"] = str(exception)
            # log_json["stack_trace"] =

        if _lib_global_do_structured_logging:
            final_log_msg = _log_msg
            if _lib_global_nested_logging_key:
                final_log_extra = {_lib_global_nested_logging_key: log_json}
            else:
                final_log_extra = log_json
        else:
            log_json["msg"] = _log_msg
            final_log_msg = json.dumps(log_json)
            final_log_extra = None

        _logger.log(level=level, msg=final_log_msg, extra=final_log_extra)

    def critical(self, **kwargs) -> None:
        return self.log(level=logging.CRITICAL, **kwargs)

    def error(self, **kwargs) -> None:
        return self.log(level=logging.ERROR, **kwargs)

    def warning(self, **kwargs) -> None:
        return self.log(level=logging.WARNING, **kwargs)

    def info(self, **kwargs) -> None:
        return self.log(level=logging.INFO, **kwargs)

    def debug(self, **kwargs) -> None:
        return self.log(level=logging.DEBUG, **kwargs)

    def log_exception(
        self,
        default_event: AuthEvent = AuthEvent.TRACE,
        override_event: AuthEvent = None,
        level: int = logging.WARNING,
        exception_cls=Exception,
        **params,
    ):
        """
        Decorator to log exceptions.
        Parameters:
            default_event : Event to log when the exception does not include a more specific event.
            override_event : Event to log regardless of whether or not the exception includes another event.
            level : Log level
            exception_cls : Exception class to catch and log.
        Example:
            ```python
            @AuthLogger.log_exception(level=logging.WARNING, event=AuthEvent.INVALID_TOKEN)
            def raise_my_exception():
                raise Some_Exception()
            ```
        """
        # some_param = params.get("some_param", default_value)

        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                try:
                    return func(*args, **kwargs)
                except exception_cls as e:
                    log_event = None
                    if override_event:
                        log_event = override_event
                    elif isinstance(e, AuthException):
                        log_event = e.event()
                    if not log_event:
                        log_event = default_event

                    self.log(event=log_event, level=level, exception=e)
                    raise e

            return wrapper

        return decorator


_default_auth_logger = AuthLogger()


# This is really only intended for use inside the library, not for use by library users.
def getAuthLogger():
    # pylint: disable=global-variable-not-assigned
    global _default_auth_logger
    return _default_auth_logger


def setPyLoggerForAuthLogger(py_logger: logging.Logger):
    """
    Set the python logger that should be used by the library.
    Parameters:
        py_logger: The python logger that the library should use.
            Set this to None to completely mute logging.
    """
    # pylint: disable=global-statement
    global _lib_global_py_logger
    _lib_global_py_logger = py_logger


def setStructuredLogging(nested_key=DEFAULT_NESTED_KEY):
    """
    Configure the library to emit structured log messages.  When this
    mode is set, logs will be emitted specifying information using the logger's
    `extra` field.

    Parameters:
        nested_key : dict key in which to wrap the library's data logged under
            the `extra` field.  The default is to include all library logged
            extra fields encapsulated inside a dictionary with the single key
            `props`.  This default was chosen to comform to the expectations of
            the `json_logging` python module. For example, by default extra data
            will be submitted to the logger as
            `log(msg="log msg", extra={"props": {library_provided_extra_keys}})`.

            Set this to `None` to forego wrapping.
    """
    # pylint: disable=global-statement
    global _lib_global_do_structured_logging
    _lib_global_do_structured_logging = True
    global _lib_global_nested_logging_key
    _lib_global_nested_logging_key = nested_key


def setStringLogging():
    """
    Configure the library to emit simple string log messages.  When this
    mode is set, logs will be emitted with all information in the log message,
    ahd the logger's `extra` field will be set to None.
    """
    # pylint: disable=global-statement
    global _lib_global_do_structured_logging
    _lib_global_do_structured_logging = False
